#ifndef MEARTHCALLBACKHANDLER_H_
#define MEARTHCALLBACKHANDLER_H_
//#include <boost/graph/graph_traits.hpp>
//#include <boost/graph/adjacency_list.hpp>
//#include <boost/graph/topological_sort.hpp>
#include <vector>
#include "CallBackHandler.h"
#include "jvmti.h"
#include "TraceOutputter.h"
#include <string>
#include <map>
#include <iostream>
#include <fstream>
#include "HashGraph.h"
#include "WeakRefCleaner.h"

using namespace std;

//using namespace boost;
//Forward declaration needed
class WeakRefCleaner;
class ETCallBackHandler : public CallBackHandler {

public:
  ETCallBackHandler (jvmtiEnv *jvmti, JavaVM* vm,  TraceOutputter* traceOutputter);
  virtual void nativeNewObj    (JNIEnv *env, jclass klass, jobject thread, jobject o, bool isWeak);
  virtual void nativeNewObj    (JNIEnv *env, jclass klass, jobject thread,
                                jint classId, jint allocSite, bool isWeak);
  virtual void nativeNewArray  (JNIEnv *env, jclass klass, jobject thread, jobject o);
  virtual void nativeMethodEntry       (JNIEnv *jni_env, jclass klass, jint methodId, jobject receiver);
  virtual void nativeMethodEntryStatic (JNIEnv *jni_env, jclass klass, jint methodId);
  virtual void nativeMethodExit ( JNIEnv *jni_env, jclass klass, jint methodId, jobject receiver);
  virtual void nativeMethodException (JNIEnv *jni_env, jclass klass,
                                      jobject exception, jint methodId, jobject receiver);
  virtual void nativeMethodExceptionStatic (JNIEnv *jni_env, jclass klass,
                                            jobject exception, jint methodId);
  virtual void nativeArrayStore (JNIEnv *jni, jclass klass,
                                 jobject array, jint index, jobject storee);
  virtual void nativePointerMovedOff (JNIEnv *env, jclass klass,
                                      jobject oldTarget, jobject origin, jobject newTarget);
  virtual void syntheticPointerMovedOff(JNIEnv *jni, jclass klass, 
      					jlong oldTargetTag,
     					jlong originTag, jlong newTargetTag);
  virtual void nativeVarEvent (JNIEnv *jni, jclass klass, jint numObjs,
                               jobject o0, jobject o1, jobject o2, jobject o3, jobject o4);
  virtual void nativeUninitPutfield (JNIEnv *jni, jclass klass,
                                     jobject newTarget, jint allocSite, jint fieldId);
  virtual void nativeInitOf (JNIEnv *jni, jclass klass, jobject thread, jint allocSite);
  virtual void ObjectFree (jvmtiEnv *jvmti, jlong tag);
  virtual void GCStart (jvmtiEnv *jvmti);
  virtual void GCEnd (jvmtiEnv *jvmti);
  virtual void shadowGCEnd (jvmtiEnv *jvmti);
  virtual void VMInit (jvmtiEnv *jvmti, JNIEnv *jni, jthread thread);
  virtual ~ETCallBackHandler ();

  jint scanLiveObject (jlong *tag_ptr);

  void weakRefCleanerThreadStart (JNIEnv* jni);

  jint scanAllObjectsCB (jlong class_tag, jlong size, jlong* tag_ptr, jint length);

  bool diedBefore (unsigned long obj1, unsigned long obj2);

private:
  map<unsigned long, unsigned long> tag2TimeStamp;
  map<string,bool> methodSet;
  map<string,long> method2id;
  bool doDeathChunks;
  //map<long, map<int, long> > arrayShadows;
  jvmtiEnv*  aJVMTI;
  vector<unsigned long> deadList;
  TraceOutputter* traceOutputter;
  jlong getTagAndTagIfNeeded (JNIEnv* jni, jobject o);

  unsigned long currentTime;
  jlong incTime () {
    return __sync_fetch_and_add(&(this->currentTime), 1) + 1;
  }

  void timeStampTag (jlong oldTargetTag);
  bool seenBefore (string method);
  //adjacency_list<hash_setS, vecS> heap;
  HashGraph heap;
  void computeObjectDeathTimes ();
  bool isUnreachable (unsigned long obj);
  bool inDeathList;
  void cleanUpLocalVarTable (jvmtiEnv* jvmti, jvmtiLocalVariableEntry *localVarTable,
                             jint entryCount);
  jlong timeStampObject (jvmtiEnv *jvmti, jobject o);
  void scanStack (jvmtiEnv *jvmti, int frame, JNIEnv *jni, jthread thread);
  jlong tagObject (jvmtiEnv *jvmti, jobject o);
  void addObjectToHeap (jlong o);
  void lockOn (jrawMonitorID lock);
  void unlockOn (jrawMonitorID lock);
  void updateHeap (jlong oldTargetTag, jlong originTag, jlong newTargetTag);
  void VMDeath (jvmtiEnv *jvmti);
  void fakeAGC (jvmtiEnv *jvmti);
  ofstream deathChunkFile;

  //Cleans up weak references that may have been nulled by the GC
  WeakRefCleaner *weakRefCleaner;

  jrawMonitorID  heapLock;
  jrawMonitorID  vertexMapLock;
  jrawMonitorID  deadListLock;
  jrawMonitorID  clockLock;
  jrawMonitorID  tagLock;
  jrawMonitorID  tag2timeStampLock;
  jrawMonitorID  weakRefMon;
  //jrawMonitorID  arrayShadowLock;

#ifdef DEBUG_MULTINODE
  vector<unsigned long> multiNodes;
  bool isMultiNode (long unsigned int);
#endif 

  //void dealWithDeath ();

};

class DiedBeforeComparator {
public:
  DiedBeforeComparator (ETCallBackHandler & etcb) : handler(etcb) {}
  
  bool operator () (unsigned long obj1, unsigned long obj2) {
    return handler.diedBefore(obj1, obj2);
  }
  
private:
  ETCallBackHandler & handler;
};

#endif /*MEARTHCALLBACKHANDLER_H_*/

// Local Variables:
// mode:C++
// c-basic-offset:2
// indent-tabs-mode:nil
// End:
