#include "ETThreadLocalData.h"
#include <stdlib.h>
#include <iostream>


using namespace std;

ETThreadLocalData::ETThreadLocalData (int threadTag) {
  this->threadTag   = threadTag;
  this->callDepth   = 0;
  this->allocInfos  = NULL;
#ifdef DEBUG_CALL_DEPTH
  this->callEntries = NULL;
#endif /* DEBUG_CALL_DEPTH */
}

int ETThreadLocalData::getCallDepth () {
  return callDepth;
}

void ETThreadLocalData::incrementCallDepth (int incr) {
  callDepth += incr;
}

void ETThreadLocalData::noteEntry (int depth, int methodId) {
#ifdef DEBUG_CALL_DEPTH
  struct callEntry *newEntry = (struct callEntry *)malloc(sizeof(struct callEntry));
  newEntry->next     = this->callEntries;
  newEntry->depth    = depth;
  newEntry->methodId = methodId;
  this->callEntries = newEntry;
#endif /* DEBUG_CALL_DEPTH */
}

void ETThreadLocalData::noteExit (int depth, int methodId) {
  struct allocInfo *ai;

  while (((ai = this->allocInfos) != NULL) && (ai->callDepth >= depth)) {
    this->allocInfos = ai;
    free(ai);
  }

#ifdef DEBUG_CALL_DEPTH
  struct callEntry *entry = this->callEntries;
  if (entry == NULL) {
    cerr << "Call depth mismatch; no entry; expected: " << depth << " " << methodId
	 << " [" << threadTag << "]" << endl;
    return;
  }
  this->callEntries = entry->next;
  if (entry->depth != depth) {
    cerr << "Call depth mismatch; is: " << entry->depth << "; expected: " << depth
	 << " [" << threadTag << "]" << endl;
  } else if (entry->methodId != methodId) {
    cerr << "Method id mismatch; is: " << entry->methodId << "; expected: " << methodId
	 << " [" << threadTag << "]" << endl;
  }
  free(entry);
#endif /* DEBUG_CALL_DEPTH */
}

void ETThreadLocalData::newObject (int allocSite, jlong newTag, unsigned long timestamp) {
  struct allocInfo *ai = (struct allocInfo *)malloc(sizeof(struct allocInfo));
  ai->next = this->allocInfos;
  ai->callDepth = this->callDepth;
  ai->allocSite = allocSite;
  ai->objectTag = newTag;
  ai->allocTime = timestamp;
  this->allocInfos = ai;
}

static struct allocInfo *getAllocInfoForSite (struct allocInfo *list, int allocSite) {
  while ((list != NULL) && (list->allocSite != allocSite)) {
    list = list->next;
  }
  return list;
}

void ETThreadLocalData::uninitPutField (int allocSite, int fieldId, jlong newValTag) {
}

void ETThreadLocalData::initOf (int allocSite) {
  struct allocInfo *ai = getAllocInfoForSite(this->allocInfos, allocSite);

  if (ai == NULL) {
    cerr << "ETThreadLocalData::initOf: allocSite " << allocSite << " not found; depth = "
         << this->callDepth << "; thread id = " << this->threadTag << endl;
    return;
  }

  struct allocInfo *newInfo = (struct allocInfo *)malloc(sizeof(struct allocInfo));
  newInfo->next = this->allocInfos;
  newInfo->callDepth = this->callDepth + 1;
  newInfo->allocSite = 0;
  newInfo->objectTag = ai->objectTag;
  newInfo->allocTime = ai->allocTime;
  this->allocInfos = newInfo;
}

ETThreadLocalData::~ETThreadLocalData () { }

// Local Variables:
// mode:C++
// c-basic-offset:2
// indent-tabs-mode:nil
// End:
