#include <HashGraph.h>
#include <set>
#include <vector>
#include <algorithm>
#include "GraphAlgorithms.h"
#include <iostream>

using namespace std;


//Return the union of sets a and b.
static set<long> setUnion(const set<long> &a, const set<long> &b){
	set<long> unionSet;

	for(set<long>::iterator it = a.begin(); it != a.end(); it++){
		unionSet.insert(*it);
	}

	for(set<long>::iterator it = b.begin(); it != b.end(); it++){
		unionSet.insert(*it);
	}
	
	return unionSet;
}


static bool containsOnly(const set<long> &s, long l){
  return (s.size() == 1 && s.count(l) == 1);
}


static bool isUnreachable(const vector<unsigned long>& deadList, unsigned long id){
  for(vector<unsigned long>::const_iterator it = deadList.begin();
      it != deadList.end();
      it++)
  {
    if(*it == id){
      return true;
    }
  }

  return false;
}


map<long, set<long> > 
computeDeathChunks(HashGraph g, vector<unsigned long> deadList, 
                   map<unsigned long, unsigned long>& tag2timeStamp)
{
    unsigned long lastTime = -1L; //This assumes two's compliment to find max long.
    map<long,set<long> > deathCause;
    
    //stack<unsigned long int> unreachableStack;
    vector<unsigned long> unreachableStack; //Need own local copy of unreachable stack
    //sort unreachable objects from the earliest timestamp to the latest

    
    GADiedBeforeComparator comp(tag2timeStamp);
    std::sort(deadList.begin(), deadList.end(), comp);//Send a function
    
    
    for(vector<unsigned long>::const_iterator it = deadList.begin(); it < deadList.end(); it++ ){
      unreachableStack.push_back(*it);
      //Initialize the death cause for each object to the object itself
      //deathCause[*it].insert(*it); 
    }





    while( !unreachableStack.empty() ){ 
      unsigned int obj = unreachableStack.back();
      unreachableStack.pop_back();
      
      unsigned long  objTime = tag2timeStamp[obj];

      
      if ( objTime <= lastTime){
              lastTime = objTime;
              
              vector<long>::iterator neighborIt;
              
              pair<vector<long>::iterator, vector<long>::iterator  > beginEnd;

              beginEnd  = g.neighbors(obj);

              for(neighborIt = beginEnd.first; neighborIt != beginEnd.second; ++neighborIt){
                      unsigned long  neighbor = *neighborIt;
                      unsigned long neighborTime = tag2timeStamp[neighbor];

                      if(isUnreachable(deadList, neighbor) && neighborTime < lastTime){


                              if(deathCause[obj].size() == 0){
                                      deathCause[neighbor].insert(obj);
                              }
                              else{
                                      deathCause[neighbor] = setUnion(deathCause[neighbor],deathCause[obj]);
                              }
                              
                              if(neighborTime < lastTime){
                                      tag2timeStamp[ neighbor ] = lastTime;
                                      unreachableStack.push_back( neighbor );	

                                      unreachableStack.push_back(neighbor);
                              }
                      }

              }
      
      }	
      
    }

  return deathCause;
}



void outPutDeathChunks(map<long,set<long> > causes,
                       ostream& deathChunkFile,
                       map<unsigned long, unsigned long>& tag2TimeStamp)
{
	
	deathChunkFile << "---Death Chunks--- " << endl;
	for(map<long,set<long> >::iterator it = causes.begin(); it != causes.end(); it++){
		long obj = it->first;
		set<long> causeList = it->second;
		
		deathChunkFile << hex <<  it->first << " (" << tag2TimeStamp[it->first] << ")" << " : ";
		for(set<long>::iterator jt = causeList.begin(); jt != causeList.end(); jt++)
                {
			deathChunkFile << (*jt) << " ";		
		}
		deathChunkFile << endl;
	}
}


