#include "WeakRefCleaner.h"
#include "ETCallBackHandler.h"
#include <vector>
#include <iostream>
#include <algorithm>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <agent_util.h>
#include <java_crw_demo.h>
#include "main.h"

using namespace std;

WeakRefCleaner::WeakRefCleaner (jvmtiEnv* jvmti, jrawMonitorID mon,  ETCallBackHandler *etcbh) {
  weakCleanMon = mon;
  etCallBackHandler = etcbh;

  jvmti->CreateRawMonitor("weakRefListLock", &weakRefListLock);	
}

jfieldID WeakRefCleaner::getReferentFieldID(jvmtiEnv* jvmti, JNIEnv* jni){
  jclass reference;
  jint fieldCount;
  jfieldID* fieldIDs;

  reference = jni->FindClass("java/lang/ref/Reference");
  jvmti->GetClassFields(reference, &fieldCount, &fieldIDs);

  for (int i = 0; i < fieldCount; i++) {
    char* fieldName;
    jvmti->GetFieldName(reference, fieldIDs[i], &fieldName, NULL, NULL);
    if (strcmp(fieldName, "referent") == 0) {
      jvmti->Deallocate((unsigned char*)fieldName);
      return fieldIDs[i];
    }
    jvmti->Deallocate((unsigned char*)fieldName);
  }
  
  cerr << "WeakRefCleaner::getReferentFiledID "
       << "could not find 'referent' field in java/lang/ref/Reference";
  exit(-1);
}

//Gets the object that has a particular tag.
//If there is more than one object it will exit with an error.
//EBM: I am concerned that this may be VERY expensive
jobject WeakRefCleaner::getObject (jvmtiEnv* jvmti, const long tag) {
  long tags[1];
  jint object_count;
  jobject* objects = NULL;
  jobject theObject;

  if (tag == 0) {
    return NULL;
  }

  tags[0] = tag;

  jvmtiError err = jvmti->GetObjectsWithTags(1, (jlong *)tags, &object_count, &objects, NULL);
  check_jvmti_error(jvmti, err, "WeakRefCleaner::getObject: could not get object from tag.");

  if (object_count > 1) {
    cerr << "WeakREfCleaner::getObject: Multiple objects found with same tag!" << endl;
    exit(1);
  }

  if (object_count == 0) {
    //Probably the object already got GC'd... nothing to do.
    return NULL;
  }

  theObject = objects[0];

  jvmti->Deallocate((unsigned char*)objects);

  return theObject;

}

static int noLongerNeighborsSub(
jvmtiHeapReferenceKind reference_kind, 
const jvmtiHeapReferenceInfo* reference_info, 
jlong class_tag, 
jlong referrer_class_tag, 
jlong size, 
jlong* tag_ptr, 
jlong* referrer_tag_ptr, 
jint length, 
void* user_data)
{
    vector<long> *confirmedNeighbors;


    confirmedNeighbors = (vector<long>*)user_data;

    confirmedNeighbors->push_back(*tag_ptr);	

    return 0; ///This means we do not visit neighbors of this object
}

/*Return a list of objects that are no longer neighbors acoording to the JVM,
but /are/ neighbors according to the shadow heap.
  Note that the list may contain the same object id's twice, if for example,
  a neighbor had multiplicity 2, and now has zero */
vector<long> WeakRefCleaner::noLongerNeighbors(jvmtiEnv* jvmti, jobject obj, HashGraph& theHeap){
  vector<long> confirmedNeighbors;
  vector<long> shadowNeighbors;
  vector<long> noLonger;
  jlong object_id;
  jvmtiError err;
  jvmtiHeapCallbacks cb;

  verbose_println("Entered no longer neighbors");

  memset(&cb,0,sizeof(jvmtiHeapCallbacks));

  cb.heap_reference_callback = noLongerNeighborsSub;

  err = jvmti->GetTag(obj,&object_id);
  check_jvmti_error(jvmti,err,"WeakRefCleaner::noLongerNeighbors: Could not get object tag");

  
  err = jvmti->FollowReferences(0, NULL, obj, &cb, (void*)&confirmedNeighbors);
  check_jvmti_error(jvmti,err, "WeakRefCleaner::noLongerNeighbors: Could not follow references.");

  pair<vector<long>::iterator, vector<long>::iterator > beginEnd = theHeap.neighbors(object_id);

  for (vector<long>::iterator it = beginEnd.first; it != beginEnd.second; it++){
    shadowNeighbors.push_back(*it);
  }
  

  //This could be potentiall very slow if some sub class of WeakRef has
  //lots of neighbors...
  for (vector<long>::iterator it = shadowNeighbors.begin();
      			      it != shadowNeighbors.end();
			      it++)
  {
      int countDif = count(shadowNeighbors.begin(), shadowNeighbors.end(), *it) 
		   - count(confirmedNeighbors.begin(), confirmedNeighbors.end(),  *it);
	
      assert(!(countDif < 0));
      if(countDif > 0){
	for(int i = 0; i < countDif; i++){
	   noLonger.push_back(*it);
	}

      }
   }
      
   return noLonger;
}




void WeakRefCleaner::cleanerThreadStart (jvmtiEnv* jvmti, JNIEnv* jni, HashGraph& theHeap) {
  jvmtiError err;

  while (true) {
    verbose_println("Top of cleaner loop.");
    err = jvmti->RawMonitorEnter(weakCleanMon);
    check_jvmti_error(jvmti, err,
                      "WeakREfCleaner::cleanerThreadStart: could not enter weakCleanMon");
    err= jvmti->RawMonitorWait(weakCleanMon,0);
    check_jvmti_error(jvmti, err,
                      "WeakRefCleaner::cleanerThreadStart: could not wait on weakCleanMon");

    set<jweak> cleanedWeakRefs;
    //Iterate over weak refs
    for (set<jweak>::iterator it = weakRefs.begin();
         it != weakRefs.end();
         it++) {

      //jobject weakRef = getObject(jvmti, *it);
      //The jweak reference may become invalid at any time,
      //Even if we test it, the test may be invalidated before we use it.
      //So, we need to create a local reference.
      //weakRef is a _strong_ local reference to a WeakReference object.
      jobject weakRef = jni->NewLocalRef(*it);

      if (jni->IsSameObject(weakRef, NULL)) {
        continue;
      }


      jlong weakRefTag;
      err  = jvmti->GetTag(weakRef, &weakRefTag);
      check_jvmti_error(jvmti, err, "WeakRefCleaner::cleanerThreadStart: could not get weak ref tag.");

      /*pair<vector<long>::iterator, vector<long>::iterator > beginEnd = theHeap.neighbors(weakRefTag);
      //Need to build a list to avoid concurrent modification

      if (jni->IsSameObject(referent, NULL)) {
        vector<long> neighbors;
        for (vector<long>::iterator neighborIt = beginEnd.first;
             neighborIt != beginEnd.second;
             ++neighborIt) {
          neighbors.push_back(*neighborIt);
        }

        //Go over all the neighbors.
        //cerr << "---Weak referece neighbors: 0x ---"<<*it << endl;
        for (vector<long>::iterator jt = neighbors.begin();
             jt != neighbors.end();
             jt++) {
          //Update pointers to null
          //long refTag = *it;
          jobject neighbor  = getObject(jvmti, *jt);
          //	cerr << hex << "0x" << *jt << endl;
          //EBM: this seems to clear *all* neighbors, but should clear only the referent field
          //  figuring this out may be challenging unless we track individual fields ...
          etCallBackHandler->nativePointerMovedOff(jni, NULL, neighbor, weakRef, NULL);
        }
        //cerr << "--------------" << endl;

        cleanedWeakRefs.insert(*it);
      }*/

      vector<long> removedNeighbors = noLongerNeighbors(jvmti,weakRef,theHeap);
      if(removedNeighbors.size() > 0){
	 for (vector<long>::iterator it = removedNeighbors.begin(); it != removedNeighbors.end(); it++){
	   etCallBackHandler->syntheticPointerMovedOff(jni,NULL,*it,weakRefTag,0);
	 }

	 cleanedWeakRefs.insert(*it);
      }


      jni->DeleteLocalRef(weakRef);
    }

    //We do this here to avoid modifying the set while we iterate over it above
    for (set<jweak>::iterator cleanRef = cleanedWeakRefs.begin();
         cleanRef != cleanedWeakRefs.end();
         cleanRef++) {
      jni->DeleteWeakGlobalRef(*cleanRef);
      weakRefs.erase(*cleanRef);
    }

    cleanedWeakRefs.clear();
    
    verbose_println("Entered WeakClean");

    jvmti->RawMonitorExit(weakCleanMon);
  }
}


void WeakRefCleaner::addWeakRef (jvmtiEnv* jvmti, jweak obj) {
  verbose_println("WeakRefCleaner::addWeakRef: Weak Ref Added to list.");
  jvmti->RawMonitorEnter(weakRefListLock);
  weakRefs.insert(obj);
  jvmti->RawMonitorExit(weakRefListLock);
}

// Local Variables:
// mode:C++
// c-basic-offset:2
// indent-tabs-mode:nil
// End:
